package com.sillycycle.bagleyd.panex;

/*
 * @(#)PanexStack.java
 *
 * Copyright 1994 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Deque;
import java.util.LinkedList;

/**
 * The <code>Panex</code> class allows one to undo and redo
 * in the <code>PanexFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class PanexStack {
	//int from, to;
	private int packed; /* This makes assumptions on the data. */
	private int count;
	private final Deque<String> stack = new LinkedList<>();

	private void writeMoveFrom(int from) {
		//this.from = from;
		packed += ((from & 0xF)  << 4);
	}

	private void writeMoveTo(int to) {
		//this.to = to;
		packed += to & 0xF;
	}

	int readMoveFrom() {
		//return this.from;
		return ((packed >> 4) & 0xF);
	}

	int readMoveTo() {
		//return this.to;
		return (packed & 0xF);
	}

	void setMove(int from, int to) {
		packed = 0;
		writeMoveFrom(from);
		writeMoveTo(to);
		stack.push(Integer.toString(packed));
		count++;
	}

	void getMove() {
		packed = Integer.parseInt(stack.pop());
		count--;
	}

	boolean madeMoves() {
		return !stack.isEmpty();
	}

	void flushMoves() {
		while (madeMoves())
			getMove();
		count = 0; // probably do not need
	}

	int numMoves() {
		return count;
	}
}
