package com.sillycycle.bagleyd.panex.view;

/*
 * @(#)PanexGeometry.java
 *
 * Copyright 2014	David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

//import android.graphics.Point;
import java.awt.Color;
import java.awt.Point;

import com.sillycycle.bagleyd.panex.model.PanexFormat;

import com.sillycycle.bagleyd.util.Colour;

/**
 * The <code>PanexGeometry</code> class holds physical info about
 * the panex widget.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/panex.html
 */

public class PanexGeometry {
	public static final int MAX_SLICES = 64;
	public static final int MULT_DELAY = 48;

	Color frameColor = Colour.tan;
	Color borderColor = Colour.brown;
	Color background = Colour.gray25;
	Color tileColor = Colour.gray75;
	Color[] pyramidColor = new Color[PanexFormat.MAX_COLUMNS - 1];
	Color slotColor = Colour.black; // used?
	Color foreground = frameColor;
	String pyramidNames;
	boolean mono = false, reverse = false;
	Point coreSize;
	Point pos;
	Point tileSize, frameSize;
	Point delta, frameOffset;
	int	 delay, numberSlices;
	PanexFormat model;

	public PanexGeometry(PanexFormat model) {
		this.model = model;
		coreSize = new Point(0, 0);
		pyramidColor[0] = Colour.magenta;
		pyramidColor[1] = Colour.orange;
	}

	public void checkGeometry() {
		if (delay < 0) {
			System.out.println(delay +
				" delay can not be negative, taking absolute value");
			delay = -delay;
		}
		if (numberSlices < 0) {
			System.out.println(numberSlices +
				" numberSlices can not be negative, taking absolute value");
			numberSlices = -numberSlices;
		}
	}

	public void resize(int width, int height) {
		/* Determine size of client area */
		coreSize = new Point(width, height);
		delta = new Point(8, 2);
		pos = new Point(java.lang.Math.max((coreSize.x -
			delta.x) / PanexFormat.MAX_COLUMNS, delta.x),
			java.lang.Math.max((coreSize.y -
			2 * delta.y - 3) / model.getSlots(), delta.y));
		frameSize = new Point(pos.x * PanexFormat.MAX_COLUMNS + delta.x + 2,
			pos.y * model.getSlots() + delta.y + 3);
		frameOffset = new Point((coreSize.x - frameSize.x +
			2) / 2, (coreSize.y - frameSize.y + 2) / 2);
		tileSize = new Point( /* Make the following even */
			((pos.x - delta.x) >> 1) << 1,
			pos.y - delta.y + 1);
	}

	public boolean getMono() {
		return mono;
	}

	public void setMono(boolean value) {
		mono = value;
	}

	public boolean getReverse() {
		return reverse;
	}

	public void setReverse(boolean value) {
		reverse = value;
	}

	public String getPyramidNames() {
		return pyramidNames;
	}

	public void setPyramidNames(String value) {
		pyramidNames = value;
	}

	public Color getForeground() {
		return foreground;
	}

	public void setForeground(Color value) {
		foreground = value;
	}

	public Color getBackground() {
		return background;
	}

	public void setBackground(Color value) {
		background = value;
	}

	public Color getFrameColor() {
		return frameColor;
	}

	public void setFrameColor(Color value) {
		frameColor = value;
	}

	public Color getBorderColor() {
		return borderColor;
	}

	public void setBorderColor(Color value) {
		borderColor = value;
	}

	public Color getPyramidColor(int index) {
		return pyramidColor[index];
	}

	public void setPyramidColor(int index, Color value) {
		pyramidColor[index] = value;
	}

	public Color getTileColor() {
		return tileColor;
	}

	public void setTileColor(Color value) {
		tileColor = value;
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int value) {
		if (value < 0) {
			delay = -value;
			return;
		}
		delay = value;
		numberSlices = ((delay < MAX_SLICES) ? delay + 1 : MAX_SLICES);
	}

	public void decrementDelay() {
		delay -= 10;
		if (delay < 0) {
			delay = 0;
		}
	}

	public void incrementDelay() {
		delay += 10;
	}

	public int getNumberSlices() {
		return numberSlices;
	}

	public void setNumberSlices(int value) {
		if (value >= 0)
			numberSlices = value;
	}

	public Point getCoreSize() {
		return coreSize;
	}

	public Point getFrameSize() {
		return frameSize;
	}

	public Point getTileSize() {
		return tileSize;
	}

	public Point getPos() {
		return pos;
	}

	public Point getDelta() {
		return delta;
	}

	public Point getFrameOffset() {
		return frameOffset;
	}

	public int cartesianX(int stack) {
		return stack * pos.x + delta.x + frameOffset.x;
	}

	public int cartesianY(int loc) {
		return loc * pos.y + delta.y + frameOffset.y;
	}

	protected String paramString() {
		//return super.paramString() +
		return "panexGeometry: " +
			",mono=" + mono +
			",reverse=" + reverse +
			",names=" + pyramidNames +
			",color0=" + pyramidColor[0] +
			",color1=" + pyramidColor[1] +
			",tileColor=" + tileColor +
			",background=" + background +
			",delay=" + delay;
	}
}
